#Final Project  - Su-Mi Hur

#import python modules
import numpy as np
import time

#import compiled fortran library
import finallib

import sys
import os

UseVisual=False
import atomwrite


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.

def CalcEnergy(Pos,PosT):
    """Wrapper for energy calculation.
Input:
    Pos: (N,3) position array
Output:
    PEnergy: potential energy, float
"""
    global M, L, Cut
    return finallib.calcenergy(Pos, PosT, M, L, Cut)


def CalcEnergyForces(Pos, Forces,PosT):
    """Wrapper for energy and force calculation.
Input:
    Pos: (N,3) position array
    Forces: (N,3) force array
Output:
    PEnergy: potential energy, float
    Forces: (N,3) force array (updated)
"""
    global M, L, Cut
    return finallib.calcenergyforces(Pos, PosT, M, L, Cut, Forces)

def VVintegrate(Pos, Vel, Accel, delt,PosT):
    """Wrapper for energy and force calculation.
Input:
    Pos: (N,3) position array
    Forces: (N,3) force array
Output:
    PEnergy: potential energy, float
    Forces: (N,3) force array (updated)
"""
    global M, L, Cut
    return finallib.vvintegrate(Pos, Vel, Accel, PosT, M, L, Cut, delt)


def LineSearch(Pos, Dir, dx, EFracTol,PosT,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = Dir / np.sqrt(np.sum(Dir * Dir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [CalcEnergy(Pos + NormDir * x,PosT) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = CalcEnergy(Pos + NormDir * dx,PosT)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [CalcEnergy(Pos + NormDir * 2. * dx,PosT)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [CalcEnergy(Pos + NormDir * Dists[-1],PosT)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = CalcEnergy(Pos + NormDir * d3,PosT)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = CalcEnergy(Pos + NormDir * d3,PosT)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG,PosT):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    PE, Forces = CalcEnergyForces(Pos, np.zeros_like(Pos),PosT)
    Dir = Forces
    OldPE = 1.e300
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS,PosT)
        OldForces = Forces.copy()
        PE, Forces = CalcEnergyForces(Pos, Forces,PosT)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def TetheredPositions(Nc,L):
    """Returns an array of tethered positions of each polymer chain,
placed randomly on a square for convenience. 
Input:
    Nc: number of polymer chains 
    L: box length
Output:
    PosT: (Nc,3) array of tethered positions
"""
    PosT = np.zeros((Nc,3), float)
    i = 0
    while i<Nc:
        x=L[0]*(np.random.random() - 0.5)
        y=L[1]*(np.random.random() - 0.5)
        PosT[i] = np.array([x,y,0.], float)
        i += 1
    return PosT


def InitPositions(N, L,PosT):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
    Lz : Domain size in z direction
Output:
    Pos: (N,3) array of positions
"""
    global M, Nc
   
    #make the position array
    Pos = np.zeros((N,3), float)

    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L[1] / NLat
    LzatSpac = L[2]/(M+4)
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L[1]
    rz = LzatSpac * np.arange(M, dtype=float) + LzatSpac# - 0.5*L[2]
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    j=0
    while j < Nc:
        i = 0
        for z in rz:
            Pos[j*M + i] = PosT[j] + np.array([0.,0.,z], float)
            #add a random offset to help initial minimization
            Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
            Pos[i] = Pos[i] + Offset
            i+=1
        j += 1
    return Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos, PosT):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #get the acceleration from the forces
    PEnergy, Accel = CalcEnergyForces(Pos, np.zeros_like(Pos), PosT)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))


def RadiusOfGyration(Pos):
    #Rg calculation :[ 1/N sum_{i} {r_i dot r_i ] - [r_man dot r_mean]
    Rg = (Pos*Pos).sum()/len(Pos)-np.dot(Pos.mean(axis=0),Pos.mean(axis=0))
    Rg = np.sqrt(Rg)
    return Rg


def Rgaveg(Pos):
    global M, N, Nc, PosT
    Rgavg=0.
    i=0
    while i< Nc:
        Rgavg+=RadiusOfGyration(Pos[i*M:(i+1)*M-1,:])
        i+=1
    return Rgavg/Nc

    
def EndPosition(Pos):
    global M, N, Nc, PosT
    PosE = []
    i=0
    while i< Nc:
        PosE.append(Pos[(i+1)*M-1,:])
        i+=1
    return np.array(PosE,float)


    
def ZDensity(Pos, L):
    Rho= np.zeros(Lz, float)
    i=0
    while i< len(Pos):
        Rho[round(Pos[i,2])]+=1
        i+=1
    return Rho




def Initialization(PosT):
    """ Initializing 
Input:
Output:
    Pos: (N,3) array of positions
    Vel: Inital velocity
    Accel: Initial Accelelation
"""
    global M, L,  T, N, Nc, delt, Cut

    #Initialization
    dx = 0.0001
    EFracTolLS = 1.e-8
    EFracTolCG = 1.e-10

    Pos = InitPositions(N, L,PosT)
    Vel = InitVelocities(N, T)
    KE = 0.5 * np.sum(Vel * Vel)
    PE=CalcEnergy(Pos,PosT)
    print "%11.4f %11.4f %11.4f" %(PE, KE, PE+KE)
    PE, Pos = ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG,PosT) 
    print "%11.4f %11.4f %11.4f" %(PE, KE, PE+KE)
    Accel = InitAccel(Pos, PosT)
    return (Pos, Vel, Accel, PE, KE)




def MDInitial():
    #Job running to find equilibrium time    
    global M, L,  T, N, Nc, delt, Cut

    t0 = 0.
    tf = 1000.
    Nstep = int(round((tf - t0)/delt + 1))
    RescaleFreq = 100

    PosT = TetheredPositions(Nc, L)
    Pos, Vel, Accel, PE, KE = Initialization(PosT)
    b = file("PE_M%d.txt" %M,"w")
    b.write("i, time, PE, KE, Etot for M%d" %M)
    
    Names = ['H']
    Names= Names + ['C' for i in range(M-2)]
    Names = Names + ['N']
    for j in Names: print j
    atomwrite.WritePdb("test.pdb", Pos, L, Names, True)

    i=0
    while i < Nstep:
        if i % RescaleFreq == 0 :  
            Vel = RescaleVelocities(Vel, T)
            atomwrite.WritePdb("test.pdb", Pos, L, Names, False)
            b.write("\n")
            b.write("%d %11.4f %11.4f %11.4f %11.4f" %(i, i*delt, PE, KE, PE+KE))
        Pos, Vel, Accel, KE, PE = VVintegrate(Pos, Vel, Accel, delt,PosT)
        print "i is %d" %i
        i +=1
        
    atomwrite.WritePdb("End.pdb", PosT, L, Names, False)
    b.close()
    print "Pos is %11.4f" %Pos[1,1]
    print "Pos is %11.4f" %Pos[2,1]
    
    return Pos, Vel, Accel, KE, PE

          
def MDrun():
    global M, L,  T, N, Nc, delt, Cut

    t0 = 0.
    te1 = 50.
    te2 = 70.
    tf = 100.
    NstepEquil1= int(round((te1 - t0)/delt + 1))
    NstepEquil2= int(round((te2 - te1)/delt + 1))
    NstepProd= int(round((tf - te2)/delt + 1))
    RescaleFreq = 100

    Names = ['H']
    Names= Names + ['C' for i in range(M-2)]
    Names = Names + ['N']
    #for j in Names: print j
    
    Nrun=5
    j=0
    while j < Nrun:
        PosT = TetheredPositions(Nc, L)
        Pos, Vel, Accel, PE, KE= Initialization(PosT)
        if j==0:
            atomwrite.WritePdb("M=%11.4f, N%11.4f and sigma=%11.4f.pdb" %(M,N,sigma), Pos, L, Names, True)
        i = 0 
        while  i < NstepEquil1:
            Pos, Vel, Accel, KE, PE = VVintegrate(Pos, Vel, Accel, delt,PosT)
            i += 1
            if i % RescaleFreq == 0 :  
                Vel = RescaleVelocities(Vel, T)
            #    if j==0:
            #        atomwrite.WritePdb("M=%11.4f, N%11.4f and sigma=%11.4f.pdb" %(M,N,sigma), Pos, L, Names, False)
        print "done equil1"
        
        i = 0
        ListE = []    
        while  i < NstepEquil2:
            Pos, Vel, Accel, KE, PE = VVintegrate(Pos, Vel, Accel, delt,PosT)
            ListE.append(KE+PE)
            i += 1
            if i % RescaleFreq == 0 :  
                Vel = RescaleVelocities(Vel, T)  #print "rescale"
            #    if j==0:
            #        atomwrite.WritePdb("M=%11.4f, N%11.4f and sigma=%11.4f.pdb" %(M,N,sigma), Pos, L, Names, False)
        ListE = np.array(ListE,float)
        E2avg = ListE.mean()

        print "done equil2"
        print "E2avg = %11.4f" %E2avg

        Vel = RescaleVelocities(Vel, 2.*(E2avg - PE)/(3.*(N-1)))
        
        time = []
        MSD=[]
        PEavg=[]
        PosE=[]
        Eavg=[]
        Rho=[]
        RhoE=[]
        i = 0
        while  i < NstepProd:
            Pos, Vel, Accel, KE, PE = VVintegrate(Pos, Vel, Accel, delt, PosT)
            i += 1
            if i % RescaleFreq==0:
                time.append(i*delt)
                Rho.append(ZDensity(Pos,L))
                PosE.append(EndPosition(Pos))
                RhoE.append(ZDensity(EndPosition(Pos),L))
                PEavg.append(PE)
                Eavg.append(PE+KE)
                if j==0:
                    atomwrite.WritePdb("M=%11.4f, N%11.4f and sigma=%11.4f.pdb" %(M,N,sigma), Pos, L, Names, False)
        
        time = np.array(time,float)
        Rho = np.array(Rho,float)
        RhoE = np.array(RhoE,float)
        PosE = np.array(PosE,float)
        PEavg=np.array(PEavg,float)
        Eavg=np.array(Eavg,float)
        if j==0:
            PosE_avg = PosE
            PE_avg = PEavg
            Rho_avg = Rho.mean(axis=0)
            RhoE_avg = RhoE.mean(axis=0)
            E_avg = Eavg
        else:
            PosE_avg += PosE
            PE_avg += PEavg
            Rho_avg += Rho.mean(axis=0)
            RhoE_avg += RhoE.mean(axis=0)
            E_avg += Eavg
        j +=1

    PosE_avg = PosE_avg/Nrun
    PE_avg = PE_avg/Nrun
    Rho_avg = Rho_avg/Nrun
    RhoE_avg = RhoE_avg/Nrun
    E_avg = E_avg/Nrun

    #Finding equailibrium MC steps
    c = file("Energy for M=%11.4f, N%11.4f and sigma=%11.4f.txt" %(M,N,sigma),"w")
    c.write("time, PE, total E, Lx Lz are %11.4f %11.4f" %(L[0],L[2]) )
    c2 = file("Rho for M=%11.4f, N%11.4f and sigma=%11.4f.txt" %(M, N,sigma),"w")
    c2.write("z, Rho, RhoE" )
    j=0
    while j<len(time):
        c.write("\n")
        c.write("%11.4f %11.4f %11.4f " %(time[j], PE_avg[j], E_avg[j]))
        j+=1
    c.close()
    j=0
    while j< len(Rho_avg):
        c2.write("\n")
        c2.write("%d %11.4f %11.4f " %(j, Rho_avg[j], RhoE_avg[j]))
        j+=1
    c2.close()




#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5
#set the random number seed; useful for debugging
#np.random.seed(342324)


#### YOU WILL NEED TO DEFINE THE FOLLOWING GLOBAL ####
#### VARIABLES FOR THE CODE ABOVE TO WORK         ####

delt = 0.001
T = 4.0



Mlist = [10, 20, 40, 60, 80, 100, 150, 200, 300]
sigmalist=[0.01, 0.02, 0.05, 0.1]


for M in Mlist:
    for sigma in sigmalist:
        #Number of polymer chains
        L=10.0
        Nc=round(L**(2.)*sigma)
        if Nc==0:
            Nc=1.0
            L=(float(Nc)/sigma)**(1./2.)
        # Number of particles
        N= M*Nc
        print "L is %11.4f" %L
        Lz= M*1.5
        L=[L,L,Lz]
        #Determining graftying location (uniform distribution)
        #MDInitial()
        MDrun()


